@@ -6,4 +6,6 @@ from .milestones import milestones
 from .slide_cors import slide_cors
 from .synchro_path import synchro_path
 from .test_fork import test_fork
+from .module_inclusion import module_inclusion
 from .critical_transition import critical_transition
+from .linearity_deviation import linearity_deviation
