@@ -26,6 +26,7 @@ def graph(
     color_cells: Union[str, None] = None,
     tips: bool = True,
     forks: bool = True,
+    nodes: Optional[List] = [],
     ax=None,
     show: Optional[bool] = None,
     save: Union[str, bool, None] = None,
@@ -49,6 +50,8 @@ def graph(
         display tip ids.
     forks
         display fork ids.
+    nodes
+        display any node id.
     ax
         Add plot to existing ax
     show
@@ -131,6 +134,17 @@ def graph(
                 textcoords="offset points",
                 bbox=bbox,
             )
+    if nodes:
+        for node in nodes:
+            ax.annotate(
+                node,
+                (proj[node, 0], proj[node, 1]),
+                ha="center",
+                va="center",
+                xytext=(-8, 8),
+                textcoords="offset points",
+                bbox=bbox,
+            )
 
     savefig_or_show("graph", show=show, save=save)
 
@@ -307,14 +321,16 @@ def trajectory(
     vals = pd.Series(
         _get_color_values(adata, color_seg, layer=layer_seg)[0], index=adata.obs_names
     )
-    R = adata.uns["graph"]["R"][~np.isnan(vals), :]  # in case of subsetted tree
+    R = pd.DataFrame(adata.uns["graph"]["R"], index=adata.uns["graph"]["cells_fitted"])
+    R = R.loc[adata.obs_names]
     vals = vals[~np.isnan(vals)]
+    R = R.loc[vals.index]
 
     def get_nval(i):
-        return np.average(vals, weights=R[:, i])
+        return np.average(vals, weights=R.loc[:, i])
 
     node_vals = np.array(list(map(get_nval, range(R.shape[1]))))
-    seg_val = node_vals[edges].mean(axis=1)
+    seg_val = node_vals[np.array(edges)].mean(axis=1)
 
     if perc_seg is not None:
         min_v, max_v = np.percentile(seg_val, perc_seg)
