@@ -3,7 +3,14 @@ from .pseudotime import pseudotime, refine_pseudotime, rename_milestones
 from .test_association import test_association
 from .fit import fit
 from .cluster import cluster
-from .correlation_tools import slide_cells, slide_cors, synchro_path
+from .correlation_tools import (
+    slide_cells,
+    slide_cors,
+    synchro_path,
+    synchro_path_multi,
+    module_inclusion,
+)
 from .correlation_tools import critical_transition, criticality_drivers
 from .bifurcation_tools import test_fork, branch_specific, activation
 from .conversion import cellrank_to_tree
+from .linearity_deviation import linearity_deviation
