@@ -19,15 +19,12 @@ from functools import reduce
 from statsmodels.stats.multitest import multipletests
 import statsmodels.formula.api as sm
 
-from scipy import sparse
-
-from joblib import delayed, Parallel
+from joblib import delayed
 from tqdm import tqdm
 
 from .. import logging as logg
 from .. import settings
-from .utils import getpath
-
+from .utils import getpath, ProgressParallel, get_X
 
 try:
     from rpy2.robjects import pandas2ri, Formula
@@ -193,28 +190,17 @@ def test_fork(
                     - brcells.loc[brcells.i == i].t.min()
                 )
 
-        if layer is None:
-            if sparse.issparse(adata.X):
-                Xgenes = adata[brcells.index, genes].X.A.T.tolist()
-            else:
-                Xgenes = adata[brcells.index, genes].X.T.tolist()
-        else:
-            if sparse.issparse(adata.layers[layer]):
-                Xgenes = adata[brcells.index, genes].layers[layer].A.T.tolist()
-            else:
-                Xgenes = adata[brcells.index, genes].layers[layer].T.tolist()
+        Xgenes = get_X(adata, brcells.index, genes, layer, togenelist=True)
 
         data = list(zip([brcells] * len(Xgenes), Xgenes))
 
-        stat = Parallel(n_jobs=n_jobs)(
-            delayed(gt_fun)(data[d])
-            for d in tqdm(
-                range(len(data)),
-                disable=n_map > 1,
-                file=sys.stdout,
-                desc="    differential expression",
-            )
-        )
+        stat = ProgressParallel(
+            n_jobs=n_jobs,
+            use_tqdm=n_map == 1,
+            total=len(data),
+            file=sys.stdout,
+            desc="    Differential expression",
+        )(delayed(gt_fun)(data[d]) for d in range(len(data)))
 
         fork_stat = fork_stat + [
             pd.DataFrame(stat, index=genes, columns=milestones + ["de_p"])
@@ -232,28 +218,18 @@ def test_fork(
 
             topgenes = topleave[topleave == leave].index
 
-            if layer is None:
-                if sparse.issparse(adata.X):
-                    Xgenes = adata[subtree.index, topgenes].X.A.T.tolist()
-                else:
-                    Xgenes = adata[subtree.index, topgenes].X.T.tolist()
-            else:
-                if sparse.issparse(adata.layers[layer]):
-                    Xgenes = adata[subtree.index, topgenes].layers[layer].A.T.tolist()
-                else:
-                    Xgenes = adata[subtree.index, topgenes].layers[layer].T.tolist()
+            Xgenes = get_X(adata, subtree.index, topgenes, layer, togenelist=True)
 
             data = list(zip([subtree] * len(Xgenes), Xgenes))
 
-            stat = Parallel(n_jobs=n_jobs)(
-                delayed(test_upreg)(data[d])
-                for d in tqdm(
-                    range(len(data)),
-                    disable=n_map > 1,
-                    file=sys.stdout,
-                    desc="    leave " + str(keys[vals == leave][0]),
-                )
-            )
+            stat = ProgressParallel(
+                n_jobs=n_jobs,
+                use_tqdm=n_map == 1,
+                total=len(data),
+                file=sys.stdout,
+                desc="    upreg " + str(keys[vals == leave][0]),
+            )(delayed(test_upreg)(data[d]) for d in range(len(data)))
+
             stat = pd.DataFrame(stat, index=topgenes, columns=["up_A", "up_p"])
             leaves_stat = leaves_stat + [stat]
 
@@ -565,29 +541,20 @@ def activation(
 
             wf = warnings.filters.copy()
             warnings.filterwarnings("ignore")
-            if layer is None:
-                if sparse.issparse(adata.X):
-                    Xgenes = adata[subtree.index, genes].X.A.T.tolist()
-                else:
-                    Xgenes = adata[subtree.index, genes].X.T.tolist()
-            else:
-                if sparse.issparse(adata.layers[layer]):
-                    Xgenes = adata[subtree.index, genes].layers[layer].A.T.tolist()
-                else:
-                    Xgenes = adata[subtree.index, genes].layers[layer].T.tolist()
+
+            Xgenes = get_X(adata, subtree.index, genes, layer, togenelist=True)
+
             warnings.filters = wf
 
             data = list(zip([subtree] * len(Xgenes), Xgenes))
 
-            acti.loc[genes] = Parallel(n_jobs=n_jobs)(
-                delayed(get_activation)(data[d])
-                for d in tqdm(
-                    range(len(data)),
-                    disable=n_map > 1,
-                    file=sys.stdout,
-                    desc="    leave " + str(keys[vals == leave][0]),
-                )
-            )
+            acti.loc[genes] = ProgressParallel(
+                n_jobs=n_jobs,
+                total=len(data),
+                use_tqdm=n_map == 1,
+                file=sys.stdout,
+                desc="    to " + str(keys[vals == leave][0]),
+            )(delayed(get_activation)(data[d]) for d in range(len(data)))
 
         allact = allact + [acti]
 
@@ -664,8 +631,11 @@ def get_activation(data):
 
     wins = pd.DataFrame(
         {
-            "start": [step * i for i in range(nwin * steps)] + subtree.t.min(),
-            "end": [step * i for i in range(nwin * steps)] + subtree.t.min() + window,
+            "start": np.array([step * i for i in range(nwin * steps)])
+            + subtree.t.min(),
+            "end": np.array([step * i for i in range(nwin * steps)])
+            + subtree.t.min()
+            + window,
         }
     )
     wins = wins.loc[wins.end < subtree.t.max()]
