@@ -1,13 +1,14 @@
-from typing import Union, Optional, List
+from typing import Union, Optional, List, Iterable
+from typing_extensions import Literal
 
 import numpy as np
 import pandas as pd
 import igraph
 from anndata import AnnData
 from scipy import sparse
+import itertools
 
-from joblib import delayed, Parallel
-from tqdm import tqdm
+from joblib import delayed
 from functools import partial
 from statsmodels.stats.weightstats import DescrStatsW
 from skmisc.loess import loess
@@ -17,7 +18,7 @@ from statsmodels.stats.multitest import multipletests
 import warnings
 from .. import logging as logg
 from .. import settings
-from .utils import getpath
+from .utils import getpath, ProgressParallel, get_X
 
 import sys
 
@@ -26,10 +27,12 @@ sys.setrecursionlimit(10000)
 from rpy2.robjects import pandas2ri, Formula
 from rpy2.robjects.packages import importr
 import rpy2.rinterface
+
 pandas2ri.activate()
 
 rmgcv = importr("mgcv")
 
+
 def slide_cells(
     adata: AnnData,
     root_milestone,
@@ -365,16 +368,7 @@ def slide_cors(
         ]
     genesets = np.concatenate([genesetA, genesetB])
 
-    if layer is None:
-        if sparse.issparse(adata.X):
-            X = adata[:, genesets].X.A
-        else:
-            X = adata[:, genesets].X
-    else:
-        if sparse.issparse(adata.layers[layer]):
-            X = adata[:, genesets].layers[layer].A
-        else:
-            X = adata[:, genesets].layers[layer]
+    X = get_X(adata, adata.obs_names, genesets, layer)
 
     X = pd.DataFrame(X, index=adata.obs_names, columns=genesets)
     X_r = X.rank(axis=0)
@@ -401,7 +395,9 @@ def slide_cors(
     corAB.columns = [str(c) for c in corAB.columns]
 
     adata.uns = uns_temp
-    adata.uns[name]["corAB"] = corAB
+    adata.uns[name]["corAB"] = dict(
+        zip(corAB.index.levels[0], [corAB.loc[l] for l in corAB.index.levels[0]])
+    )
 
     logg.hint(
         "added \n" "    .uns['" + name + "']['corAB'], gene-gene correlation modules."
@@ -414,6 +410,8 @@ def synchro_path(
     adata: AnnData,
     root_milestone,
     milestones,
+    genesetA: Optional[Iterable] = None,
+    genesetB: Optional[Iterable] = None,
     n_map=1,
     n_jobs=None,
     layer: Optional[str] = None,
@@ -494,7 +492,12 @@ def synchro_path(
 
     name = root_milestone + "->" + "<>".join(milestones)
 
-    bif = adata.uns[name]["fork"]
+    if genesetA is None:
+        bif = adata.uns[name]["fork"]
+        genesetA = bif.index[(bif.module == "early") & (bif.branch == milestones[0])]
+        genesetB = bif.index[(bif.module == "early") & (bif.branch == milestones[1])]
+
+    genesets = np.concatenate([genesetA, genesetB])
 
     if n_map == 1:
         logg.info("    single mapping")
@@ -508,37 +511,12 @@ def synchro_path(
         )
         img.add_edges(edges)
 
-        genesetA = bif.index[(bif.module == "early") & (bif.branch == milestones[0])]
-        genesetB = bif.index[(bif.module == "early") & (bif.branch == milestones[1])]
-        genesets = np.concatenate([genesetA, genesetB])
-
         def synchro_milestone(leave):
-            cells = getpath(img, root, graph["tips"], leave, graph, df).index
+            cells = getpath(img, root, graph["tips"], leave, graph, df)
+            cells = cells.sort_values("t").index
 
-            if layer is None:
-                if sparse.issparse(adata.X):
-                    mat = pd.DataFrame(
-                        adata[cells, genesets].X.A, index=cells, columns=genesets
-                    )
-                else:
-                    mat = pd.DataFrame(
-                        adata[cells, genesets].X, index=cells, columns=genesets
-                    )
-            else:
-                if sparse.issparse(adata.layers[layer]):
-                    mat = pd.DataFrame(
-                        adata[cells, genesets].layers[layer].A,
-                        index=cells,
-                        columns=genesets,
-                    )
-                else:
-                    mat = pd.DataFrame(
-                        adata[cells, genesets].layers[layer],
-                        index=cells,
-                        columns=genesets,
-                    )
-
-            mat = mat.iloc[adata.obs.t[mat.index].argsort().values, :]
+            X = get_X(adata, cells, genesets, layer)
+            mat = pd.DataFrame(X, index=cells, columns=genesets)
 
             if permut == True:
                 winperm = np.min([winp, mat.shape[0]])
@@ -577,35 +555,35 @@ def synchro_path(
                     }
                 )
 
-            return pd.concat(
-                list(
-                    map(
-                        slide_path,
-                        tqdm(
-                            np.arange(0, mat.shape[0] - w, step),
-                            disable=n_map > 1,
-                            desc="    leave " + str(keys[vals == leave][0]),
-                        ),
-                    )
-                ),
-                axis=1,
-            ).T
+            ww = np.arange(0, mat.shape[0] - w, step)
+
+            res = ProgressParallel(
+                n_jobs=n_jobs,
+                total=len(ww),
+                use_tqdm=n_map == 1,
+                file=sys.stdout,
+                desc="    to " + str(keys[vals == leave][0]),
+            )(delayed(slide_path)(i) for i in ww)
+
+            return pd.concat(res, axis=1).T
 
         return pd.concat(list(map(synchro_milestone, leaves)), keys=milestones)
 
     if n_map > 1:
         permut = False
-        stats = Parallel(n_jobs=n_jobs)(
-            delayed(synchro_map)(i)
-            for i in tqdm(range(n_map), desc="    multi mapping ")
-        )
+        stats = ProgressParallel(
+            n_jobs=n_jobs, total=n_map, file=sys.stdout, desc="    multi mapping"
+        )(delayed(synchro_map)(i) for i in range(n_map))
         allcor_r = pd.concat(stats)
         if perm:
             permut = True
-            stats = Parallel(n_jobs=n_jobs)(
-                delayed(synchro_map)(i)
-                for i in tqdm(range(n_map), desc="    multi mapping permutations")
-            )
+
+            stats = ProgressParallel(
+                n_jobs=n_jobs,
+                total=n_map,
+                file=sys.stdout,
+                desc="    multi mapping permutations",
+            )(delayed(synchro_map)(i) for i in range(n_map))
             allcor_p = pd.concat(stats)
             allcor = pd.concat([allcor_r, allcor_p], keys=["real", "permuted"])
         else:
@@ -683,14 +661,35 @@ def synchro_path(
 
     adata.uns = uns_temp
 
-    adata.uns[name]["synchro"] = allcor
+    allcor = dict(
+        zip(
+            allcor.index.levels[0],
+            [
+                dict(
+                    zip(
+                        allcor.loc[l1].index.levels[0],
+                        [
+                            allcor.loc[l1].loc[l2]
+                            for l2 in allcor.loc[l1].index.levels[0]
+                        ],
+                    )
+                )
+                for l1 in allcor.index.levels[0]
+            ],
+        )
+    )
+
+    if name in adata.uns:
+        adata.uns[name]["synchro"] = allcor
+    else:
+        adata.uns[name] = {"synchro": allcor}
 
     logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
     logg.hint(
         "added \n"
         "    .uns['"
         + name
-        + "']['syncho'], mean local gene-gene correlations of all possible gene pairs inside one module, or between the two modules.\n"
+        + "']['synchro'], mean local gene-gene correlations of all possible gene pairs inside one module, or between the two modules.\n"
         "    .obs['inter_cor "
         + name
         + "'], loess fit of inter-module mean local gene-gene correlations prior to bifurcation."
@@ -699,6 +698,448 @@ def synchro_path(
     return adata if copy else None
 
 
+def synchro_path_multi(
+    adata: AnnData, root_milestone, milestones, copy=False, **kwargs
+):
+    """\
+    Estimates pseudotime trends of local intra- and inter-module correlations of fates-specific modules.
+
+    Parameters
+    ----------
+    adata
+        Annotated data matrix.
+    root_milestone
+        tip defining progenitor branch.
+    milestones
+        tips defining the progenies branches.
+    kwargs
+        arguments to pass to tl.synchro_path.
+
+    Returns
+    -------
+    adata : anndata.AnnData
+        if `copy=True` it returns subsetted or else subset (keeping only
+        significant features) and add fields to `adata`:
+
+        `.uns['root_milestone->milestoneA<>milestoneB']['synchro']`
+            Dataframe containing mean local gene-gene correlations of all possible gene pairs inside one module, or between the two modules.
+        `.obs['intercor root_milestone->milestoneA<>milestoneB']`
+            loess fit of inter-module mean local gene-gene correlations prior to bifurcation
+
+    """
+
+    adata = adata.copy() if copy else adata
+
+    logg.info("computing local correlations", reset=True)
+
+    graph = adata.uns["graph"]
+
+    edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
+    img = igraph.Graph()
+    img.add_vertices(
+        np.unique(graph["pp_seg"][["from", "to"]].values.flatten().astype(str))
+    )
+    img.add_edges(edges)
+
+    uns_temp = adata.uns.copy()
+
+    if "milestones_colors" in adata.uns:
+        mlsc = adata.uns["milestones_colors"].copy()
+
+    dct = graph["milestones"]
+    keys = np.array(list(dct.keys()))
+    vals = np.array(list(dct.values()))
+
+    leaves = list(map(lambda leave: dct[leave], milestones))
+    root = dct[root_milestone]
+
+    name = root_milestone + "->" + "<>".join(milestones)
+
+    bif = adata.uns[name]["fork"]
+
+    genesets = dict(
+        zip(
+            milestones,
+            [
+                bif.index[(bif.module == "early") & (bif.branch == m)]
+                for m in milestones
+            ],
+        )
+    )
+
+    pairs = list(itertools.combinations(milestones, 2))
+
+    for m_pair in pairs:
+        synchro_path(
+            adata,
+            root_milestone,
+            m_pair,
+            genesetA=genesets[m_pair[0]],
+            genesetB=genesets[m_pair[1]],
+            **kwargs
+        )
+
+
+def module_inclusion(
+    adata,
+    root_milestone,
+    milestones,
+    w: int = 300,
+    step: int = 30,
+    pseudotime_offset: Union["all", float] = "all",
+    module: Literal["all", "early"] = "all",
+    n_perm: int = 10,
+    n_map: int = 1,
+    map_cutoff: float = 0.8,
+    n_jobs: int = 1,
+    alp: int = 10,
+    autocor_cut: float = 0.95,
+    iterations: int = 15,
+    parallel_mode: Literal["window", "mappings"] = "window",
+    identify_early_features: bool = True,
+    layer=None,
+    perm: bool = False,
+    winp: int = 10,
+    copy: bool = False,
+):
+    """\
+    Estimates the pseudotime onset of a feature within its fate-specific module.
+
+    Parameters
+    ----------
+    adata
+        Annotated data matrix.
+    root_milestone
+        tip defining progenitor branch.
+    milestones
+        tips defining the progenies branches.
+    w
+        local window, in number of cells, to estimate correlations.
+    step
+        steps, in number of cells, between local windows.
+    pseudotime_offset
+        restrict the cell selection up to a pseudotime offset after the fork
+    module
+        restrict the gene selection to already classified early genes.
+    n_perm
+        number of permutations used to estimate the background local correlations.
+    n_map
+        number of probabilistic cells projection to use for estimates.
+    map_cutoff
+        proportion of mapping in which inclusion pseudotimne was found for a given feature to keep it.
+    n_jobs
+        number of cpu processes to perform estimates.
+    alp
+        parameter regulating stringency of inclusion event.
+    autocor_cut
+        cutoff on correlation of inclusion times between sequential iterations of the algorithm to stop it.
+    iterations
+        maximum number of iterations of the algorithm.
+    parallel_mode
+        whether to run in parallel over the windows of cells or the mappings.
+    identify_early_features
+        classify a feature as early if its inclusion pseudotime is before the bifurcation
+    layer
+        adata layer to use for estimates.
+    perm
+        do local estimates for locally permuted expression matrix.
+    winp
+        window of permutation in cells.
+    copy
+        Return a copy instead of writing to adata.
+
+    Returns
+    -------
+    adata : anndata.AnnData
+        if `copy=True` it returns subsetted or else subset (keeping only
+        significant features) and add fields to `adata`:
+
+        `.uns['root_milestone->milestoneA<>milestoneB']['module_inclusion']`
+            Dataframes ontaining inclusion timing for each gene (rows) in each probabilistic cells projection (columns).
+        `.uns['root_milestone->milestoneA<>milestoneB']['fork']`
+            Updated with 'inclusion' pseudotime column and 'module column if `identify_early_features=True`'
+    """
+
+    adata = adata.copy() if copy else adata
+
+    logg.info("Calculating onset of features within their own module", reset=True)
+
+    graph = adata.uns["graph"]
+
+    edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
+    img = igraph.Graph()
+    img.add_vertices(
+        np.unique(graph["pp_seg"][["from", "to"]].values.flatten().astype(str))
+    )
+    img.add_edges(edges)
+
+    uns_temp = adata.uns.copy()
+
+    if "milestones_colors" in adata.uns:
+        mlsc = adata.uns["milestones_colors"].copy()
+
+    dct = graph["milestones"]
+    keys = np.array(list(dct.keys()))
+    vals = np.array(list(dct.values()))
+
+    leaves = list(map(lambda leave: dct[leave], milestones))
+    root = dct[root_milestone]
+
+    name = root_milestone + "->" + "<>".join(milestones)
+
+    common_seg = list(
+        set.intersection(
+            *list(
+                map(
+                    lambda l: set(img.get_shortest_paths(str(root), str(l))[0]),
+                    leaves,
+                )
+            )
+        )
+    )
+    common_seg = np.array(img.vs["name"], dtype=int)[common_seg]
+    fork_t = adata.uns["graph"]["pp_info"].loc[common_seg, "time"].max()
+
+    def onset_map(m):
+        df = adata.uns["pseudotime_list"][str(m)]
+        edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
+        img = igraph.Graph()
+        img.add_vertices(
+            np.unique(graph["pp_seg"][["from", "to"]].values.flatten().astype(str))
+        )
+        img.add_edges(edges)
+
+        def onset_milestone(milestone):
+            sel = adata.uns[name]["fork"].branch == milestone
+            sel = (
+                sel & (adata.uns[name]["fork"].module == "early")
+                if module == "early"
+                else sel
+            )
+            geneset = adata.uns[name]["fork"].index[
+                adata.uns[name]["fork"].branch == milestone
+            ]
+            cells = getpath(
+                img,
+                root,
+                graph["tips"],
+                adata.uns["graph"]["milestones"][milestone],
+                graph,
+                df,
+            )
+
+            if pseudotime_offset != "all":
+                cells = cells.loc[cells.t < (fork_t + pseudotime_offset)]
+
+            cells = cells.sort_values("t").index
+
+            X = get_X(adata, cells, geneset, layer)
+            mat = pd.DataFrame(X, index=cells, columns=geneset)
+
+            if perm:
+                winperm = np.min([winp, mat.shape[0]])
+                for i in np.arange(0, mat.shape[0] - winperm, winperm):
+                    mat.iloc[i : (i + winperm), :] = (
+                        mat.iloc[i : (i + winperm), :]
+                        .apply(np.random.permutation, axis=0)
+                        .values
+                    )
+
+            mat = mat.T
+            ww = np.arange(0, len(cells) - w, step)
+            logW = pd.DataFrame(1, index=geneset, columns=range(len(ww)))
+
+            def slide_cor(i):
+                cls0 = mat.columns[ww[i] : min(ww[i] + w - 1, len(cells))]
+                mat_1 = mat[cls0]
+                mat_2 = mat_1.copy()
+                mat_2[mat_2 != 0] = 1
+                mat_2 = mat_2.dot(mat_2.T)
+                mat_2[mat_2 < 10] = 0
+                mat_2[mat_2 >= 10] = 1
+                cor = mat_1.T.corr(method="spearman") * mat_2
+
+                def perm_mat():
+                    mat_perm = mat[cls0].apply(np.random.permutation, axis=0).values
+                    mat_perm = pd.DataFrame(mat_perm, columns=cls0)
+                    mat_perm_2 = mat_perm.copy()
+                    mat_perm_2[mat_perm_2 != 0] = 1
+                    mat_perm_2 = mat_perm_2.dot(mat_perm_2.T)
+                    mat_perm_2[mat_perm_2 < 10] = 0
+                    mat_perm_2[mat_perm_2 >= 10] = 1
+                    return mat_perm.T.corr(method="spearman") * mat_perm_2
+
+                allperm = [perm_mat() for i in range(n_perm)]
+
+                return cor, allperm
+
+            res = ProgressParallel(
+                n_jobs=n_jobs,
+                total=len(ww),
+                use_tqdm=n_map == 1,
+                file=sys.stdout,
+                desc="    to " + milestone,
+            )(delayed(slide_cor)(i) for i in range(len(ww)))
+
+            cors = [r[0] for r in res]
+            cors_p = [r[1] for r in res]
+
+            def corEvol(cors, cors_p, logW):
+                cor_Ps = []
+                for i in range(len(cors)):
+                    cor, cor_p = cors[i], cors_p[i]
+                    corTrue = (cor.T * logW.iloc[:, i]).mean(axis=0)
+                    cor_control = [
+                        (cor_p.T.values * logW.iloc[:, i].values).mean(axis=0)
+                        for cor_p in cor_p
+                    ]
+                    cor_Ps = cor_Ps + [
+                        np.array(
+                            [
+                                sum(
+                                    cor_c + np.random.uniform(0, 0.01, len(cor_c))
+                                    >= corTrue
+                                    for cor_c in cor_control
+                                )
+                            ]
+                        )
+                        / len(cor_control)
+                    ]
+                return cor_Ps
+
+            logList = []
+            autocor = 0
+
+            def switch_point(r, alp):
+                def logL(j):
+                    v = 0
+                    if j >= 2:
+                        v = v
+                    if j <= len(r):
+                        v = v + sum(
+                            -alp * r[j : len(r)] + np.log(alp / (1 - np.exp(-alp)))
+                        )
+
+                    return v
+
+                switch_point = np.argmax([logL(j) for j in range(len(r) + 1)])
+                n = len([logL(j) for j in range(len(r) + 1)])
+                return np.concatenate(
+                    [np.repeat(0, switch_point), np.repeat(1, n - switch_point - 1)]
+                )
+
+            i = 0
+            auto_cor = 0
+            while (autocor < autocor_cut) & (i <= iterations):
+                pMat = np.vstack(corEvol(cors, cors_p, logW)).T
+                sws = [switch_point(pMat[i, :], alp) for i in range(pMat.shape[0])]
+                sws = np.vstack(sws)
+                logW = pd.DataFrame(sws, index=geneset)
+                logList = logList + [logW]
+
+                if i > 0:
+                    autocor = np.corrcoef(
+                        logList[i - 1].idxmax(axis=1), logList[i].idxmax(axis=1)
+                    )[1][0]
+
+                i = i + 1
+
+            incl_t = pd.Series(np.nan, index=logW.index)
+            logW = logW.loc[logW.sum(axis=1) != 0]
+
+            incl_t[logW.index] = [
+                adata.obs.t[cells[(pos * (step - 1)) : (pos * (step - 1) + w)]].mean()
+                for pos in logW.idxmax(axis=1)
+            ]
+            return incl_t
+
+        return [onset_milestone(milestone) for milestone in milestones]
+
+    n_jobs_map = 1
+    if parallel_mode == "mappings":
+        n_jobs_map = n_jobs
+        n_jobs = 1
+
+    stats = ProgressParallel(
+        n_jobs=n_jobs_map,
+        total=n_map,
+        use_tqdm=n_map > 1,
+        file=sys.stdout,
+        desc="    multi mapping",
+    )(delayed(onset_map)(i) for i in range(n_map))
+
+    matSwitch = dict(
+        zip(
+            milestones,
+            [
+                pd.concat(
+                    [s[i] for s in stats],
+                    axis=1,
+                    keys=np.arange(len([s[i] for s in stats])).astype(str),
+                )
+                for i in range(len(leaves))
+            ],
+        )
+    )
+
+    perm_str = "_perm" if perm else ""
+    if perm:
+        identify_early_features = False
+
+    for m in milestones:
+        props = 1 - np.isnan(matSwitch[m]).sum(axis=1) / n_map
+        matSwitch[m] = matSwitch[m].loc[props > map_cutoff]
+
+    adata.uns[name]["module_inclusion" + perm_str] = matSwitch
+
+    updated = "."
+    if perm == False:
+        common_seg = list(
+            set.intersection(
+                *list(
+                    map(
+                        lambda l: set(img.get_shortest_paths(str(root), str(l))[0]),
+                        leaves,
+                    )
+                )
+            )
+        )
+        common_seg = np.array(img.vs["name"], dtype=int)[common_seg]
+        fork_t = adata.uns["graph"]["pp_info"].loc[common_seg, "time"].max()
+
+        g_early = []
+        dfs = []
+        for i in range(len(leaves)):
+            included = pd.concat([s[i] for s in stats], axis=1).mean(axis=1)
+            df = included[~np.isnan(included)].sort_values()
+            adata.uns[name]["fork"].loc[df.index, "inclusion"] = df.values
+            g_early = g_early + [df.index[df < fork_t]]
+            dfs = dfs + [df]
+
+        if identify_early_features:
+            updated = " and 'module'."
+            adata.uns[name]["fork"]["module"] = np.nan
+            for df, g_e in zip(dfs, g_early):
+                adata.uns[name]["fork"].loc[df.index, "module"] = "late"
+                adata.uns[name]["fork"].loc[g_e, "module"] = "early"
+
+    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
+    logg.hint(
+        "added \n"
+        "    .uns['"
+        + name
+        + "']['module_inclusion"
+        + perm_str
+        + "'], milestone specific dataframes containing inclusion timing for each gene in each probabilistic cells projection.\n"
+        + "    .uns['"
+        + name
+        + "']['fork'] has been updated with the column 'inclusion'"
+        + updated
+    )
+
+    return adata if copy else None
+
+
 def critical_transition(
     adata: AnnData,
     root_milestone,
@@ -709,7 +1150,8 @@ def critical_transition(
     w=100,
     step=30,
     loess_span=0.4,
-    gamma = 1.5,
+    gamma=1.5,
+    n_points=200,
     copy: bool = False,
 ):
     """\
@@ -798,7 +1240,8 @@ def critical_transition(
     root = dct[root_milestone]
 
     name = root_milestone + "->" + "<>".join(milestones)
-    def critical_map(m,gamma,loess_span):
+
+    def critical_map(m, gamma, loess_span):
         df = adata.uns["pseudotime_list"][str(m)]
         edges = graph["pp_seg"][["from", "to"]].astype(str).apply(tuple, axis=1).values
         img = igraph.Graph()
@@ -810,28 +1253,8 @@ def critical_transition(
         def critical_milestone(leave):
             cells = getpath(img, root, graph["tips"], leave, graph, df).index
 
-            if layer is None:
-                if sparse.issparse(adata.X):
-                    mat = pd.DataFrame(
-                        adata[cells].X.A, index=cells, columns=adata.var_names
-                    )
-                else:
-                    mat = pd.DataFrame(
-                        adata[cells].X, index=cells, columns=adata.var_names
-                    )
-            else:
-                if sparse.issparse(adata.layers[layer]):
-                    mat = pd.DataFrame(
-                        adata[cells].layers[layer].A,
-                        index=cells,
-                        columns=adata.var_names,
-                    )
-                else:
-                    mat = pd.DataFrame(
-                        adata[cells].layers[layer],
-                        index=cells,
-                        columns=adata.var_names,
-                    )
+            X = get_X(adata, cells, adata.var_names, layer)
+            mat = pd.DataFrame(X, index=cells, columns=adata.var_names)
 
             mat = mat.iloc[adata.obs.t[mat.index].argsort().values, :]
 
@@ -845,20 +1268,20 @@ def critical_transition(
                 R_cell = np.nanmean(
                     np.abs(cor_cell[np.triu_indices(cor_cell.shape[0], k=1)])
                 )
-                return [adata.obs.t[cls].max(), R_gene / R_cell,cls]
+                return [adata.obs.t[cls].mean(), R_gene / R_cell, cls]
 
-            stats = Parallel(n_jobs=n_jobs)(
-                delayed(slide_path)(i)
-                for i in tqdm(
-                    np.arange(0, mat.shape[0] - w, step),
-                    disable=n_map > 1,
-                    file=sys.stdout,
-                    desc="    leave " + str(keys[vals == leave][0]),
-                )
-            )
+            wins = np.arange(0, mat.shape[0] - w, step)
+
+            stats = ProgressParallel(
+                n_jobs=n_jobs,
+                total=len(wins),
+                use_tqdm=n_map == 1,
+                file=sys.stdout,
+                desc="    to " + str(keys[vals == leave][0]),
+            )(delayed(slide_path)(i) for i in wins)
 
             cells_l = [s[2] for s in stats]
-            stats = pd.DataFrame([[s[0],s[1]] for s in stats], columns=("t", "ci"))
+            stats = pd.DataFrame([[s[0], s[1]] for s in stats], columns=("t", "ci"))
 
             l = loess(stats.t, stats.ci, span=loess_span)
             l.fit()
@@ -869,48 +1292,66 @@ def critical_transition(
             stats["ll"] = conf.lower
             stats["ul"] = conf.upper
 
-            cell_stats=[pd.DataFrame(np.repeat(stats.ci[i].reshape(-1, 1),
-                                               len(cells_l[i])),
-                                     index=cells_l[i],
-                                     columns=["ci"]) for i in range(stats.shape[0])]
+            cell_stats = [
+                pd.DataFrame(
+                    np.repeat(stats.ci[i].reshape(-1, 1), len(cells_l[i])),
+                    index=cells_l[i],
+                    columns=["ci"],
+                )
+                for i in range(stats.shape[0])
+            ]
 
             cell_stats = pd.concat(cell_stats, axis=1)
             cell_stats = cell_stats.T.groupby(level=0).mean().T
-            cell_stats["t"]=adata.obs.loc[cell_stats.index,"t"]
-            global rmgcv
-            m = rmgcv.gam(
-                Formula("ci~s(t,bs='ts')"),
-                data=cell_stats,
-                gamma=gamma,
+            cell_stats["t"] = adata.obs.loc[cell_stats.index, "t"]
+
+            l = loess(cell_stats.t, cell_stats.ci, span=loess_span)
+            pred = l.predict(cell_stats.t, stderror=True)
+
+            cell_stats["fit"] = pred.values
+
+            lspaced_stats = pd.DataFrame(
+                {
+                    "t": np.linspace(
+                        cell_stats["t"].min(), cell_stats["t"].max(), n_points
+                    )
+                }
             )
+            pred = l.predict(lspaced_stats.t, stderror=True)
+            lspaced_stats["fit"] = pred.values
 
-            cell_stats["fit"]=rmgcv.predict_gam(m)
             del cell_stats["t"]
-            return stats,cell_stats
+            return stats, cell_stats, lspaced_stats
 
         res = list(map(critical_milestone, leaves))
 
         cell_stats = pd.concat([r[1] for r in res]).groupby(level=0).mean()
 
-        res_slide = dict(zip(milestones,[r[0] for r in res]))
+        res_slide = dict(zip(milestones, [r[0] for r in res]))
 
-        return cell_stats, res_slide
+        res_lspaced = dict(zip(milestones, [r[2] for r in res]))
+
+        return cell_stats, res_slide, res_lspaced
 
     if n_map == 1:
-        df,res_slide = critical_map(0,gamma,loess_span)
+        df, res_slide, res_lspaced = critical_map(0, gamma, loess_span)
     else:
-        #TODO: adapt multimapping
+        # TODO: adapt multimapping
         stats = Parallel(n_jobs=n_jobs)(
             delayed(critical_map)(i)
             for i in tqdm(range(n_map), file=sys.stdout, desc="    multi mapping ")
         )
         res_slides = pd.concat(stats)
 
-
     if name in adata.uns:
-        adata.uns[name]["critical transition"] = res_slide
+        adata.uns[name]["critical transition"] = {
+            "LOESS": res_slide,
+            "eLOESS": res_lspaced,
+        }
     else:
-        adata.uns[name] = {"critical transition": res_slide}
+        adata.uns[name] = {
+            "critical transition": {"LOESS": res_slide, "eLOESS": res_lspaced}
+        }
 
     adata.obs.loc[df.index, name + " CI"] = df.ci.values
 
@@ -922,8 +1363,12 @@ def critical_transition(
         "    .uns['"
         + name
         + "']['critical transition'], df containing local critical transition index per window of cells.\n"
-        "    .obs['"+ name+ " CI'], local critical transition index projected onto cells.\n"
-        "    .obs['"+ name+ " CI fitted'], GAM fit of local critical transition index projected onto cells."
+        "    .obs['"
+        + name
+        + " CI'], local critical transition index projected onto cells.\n"
+        "    .obs['"
+        + name
+        + " CI fitted'], GAM fit of local critical transition index projected onto cells."
     )
 
     return adata if copy else None
@@ -933,6 +1378,8 @@ def criticality_drivers(
     adata: AnnData,
     root_milestone,
     milestones,
+    t_span=None,
+    confidence_level: float = 0.95,
     layer: Optional[str] = None,
     device="cpu",
     copy: bool = False,
@@ -941,6 +1388,8 @@ def criticality_drivers(
     """\
     Calculates correlations between genes and local critical transition index along trajectory.
 
+    Fisher test for the correlations comes from CellRank function `cr.tl.lineages_drivers`.
+
     Parameters
     ----------
     adata
@@ -949,6 +1398,10 @@ def criticality_drivers(
         tip defining progenitor branch.
     milestones
         tips defining the progenies branches.
+    t_span
+        restrict correlations to a window of pseudotime
+    confidence_level
+        correlation confidence interval.
     layer
         adata layer to use for estimates.
     device
@@ -977,13 +1430,21 @@ def criticality_drivers(
     name = root_milestone + "->" + "<>".join(milestones)
     obs_name = name + " CI fitted"
 
-    X = adata[~np.isnan(adata.obs[obs_name])].X
-    CI = adata[~np.isnan(adata.obs[obs_name])].obs[obs_name].values
+    if t_span is None:
+        cells = adata.obs_names[~np.isnan(adata.obs[obs_name])]
+    else:
+        cells = adata.obs_names[
+            (~np.isnan(adata.obs[obs_name]))
+            & (adata.obs.t > t_span[0])
+            & (adata.obs.t < t_span[1])
+        ]
+
+    CI = adata[cells].obs[obs_name].values
 
     if layer is None:
-        X = adata[~np.isnan(adata.obs[obs_name])].X
+        X = adata[cells].X
     else:
-        X = adata[~np.isnan(adata.obs[obs_name])].layers[layer]
+        X = adata[cells].layers[layer]
 
     if device == "cpu":
         from .utils import cor_mat_cpu
@@ -998,7 +1459,10 @@ def criticality_drivers(
         X = csr_gpu(X) if sparse.issparse(X) else cp.array(X)
         corr = cor_mat_gpu(X, cp.array(CI).reshape(-1, 1)).ravel().get()
 
-    # fisher testing of correlations, code borrowed from CellRank :)
+    ### Fisher testing of correlations, CellRank implementation
+    ### https://github.com/theislab/cellrank/blob/b6345d5e6dd148317782ffc9a9f96793ad98ead9/cellrank/tl/_utils.py#L488
+    ### Copyright (c) 2019, Theis Lab
+
     confidence_level = 0.95
     n = adata.shape[0]
     ql = 1 - confidence_level - (1 - confidence_level) / 2.0
@@ -1012,6 +1476,8 @@ def criticality_drivers(
     corr_ci_high = np.tanh(mean + z * se)
     pvals = 2 * norm.cdf(-np.abs(z_score))
 
+    ###
+
     res = pd.DataFrame(
         {"corr": corr, "pval": pvals, "ci_low": corr_ci_low, "ci_high": corr_ci_high},
         index=adata.var_names,
@@ -1022,7 +1488,9 @@ def criticality_drivers(
         res[~np.isnan(pvals)].pval.values, alpha=0.05, method="fdr_bh"
     )[1]
 
-    adata.uns[name]["criticality drivers"] = res.sort_values("corr", ascending=False).dropna()
+    adata.uns[name]["criticality drivers"] = res.sort_values(
+        "corr", ascending=False
+    ).dropna()
 
     logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
     logg.hint(
