@@ -338,33 +338,8 @@ def tree_ppt(
     seed: Optional[int] = None,
 ):
 
-    if use_rep not in adata.obsm.keys() and f"X_{use_rep}" in adata.obsm.keys():
-        use_rep = f"X_{use_rep}"
+    X = get_data(adata, use_rep, ndims_rep)
 
-    if (
-        (use_rep not in adata.layers.keys())
-        & (use_rep not in adata.obsm.keys())
-        & (use_rep != "X")
-    ):
-        use_rep = "X" if adata.n_vars < 50 or n_pcs == 0 else "X_pca"
-        n_pcs = None if use_rep == "X" else n_pcs
-
-    if use_rep == "X":
-        ndims_rep = None
-        if sparse.issparse(adata.X):
-            X = DataFrame(adata.X.A, index=adata.obs_names)
-        else:
-            X = DataFrame(adata.X, index=adata.obs_names)
-    elif use_rep in adata.layers.keys():
-        if sparse.issparse(adata.layers[use_rep]):
-            X = DataFrame(adata.layers[use_rep].A, index=adata.obs_names)
-        else:
-            X = DataFrame(adata.layers[use_rep], index=adata.obs_names)
-    elif use_rep in adata.obsm.keys():
-        X = DataFrame(adata.obsm[use_rep], index=adata.obs_names)
-
-    if ndims_rep is not None:
-        X = X.iloc[:, :ndims_rep]
     X_t = X.values.T
 
     # if seed is not None:
@@ -615,32 +590,7 @@ def tree_epg(
     verbose: bool = True,
 ):
 
-    if use_rep not in adata.obsm.keys() and f"X_{use_rep}" in adata.obsm.keys():
-        use_rep = f"X_{use_rep}"
-
-    if (
-        (use_rep not in adata.layers.keys())
-        & (use_rep not in adata.obsm.keys())
-        & (use_rep != "X")
-    ):
-        use_rep = "X" if adata.n_vars < 50 or n_pcs == 0 else "X_pca"
-        n_pcs = None if use_rep == "X" else n_pcs
-
-    if use_rep == "X":
-        if sparse.issparse(adata.X):
-            X = DataFrame(adata.X.A, index=adata.obs_names)
-        else:
-            X = DataFrame(adata.X, index=adata.obs_names)
-    elif use_rep in adata.layers.keys():
-        if sparse.issparse(adata.layers[use_rep]):
-            X = DataFrame(adata.layers[use_rep].A, index=adata.obs_names)
-        else:
-            X = DataFrame(adata.layers[use_rep], index=adata.obs_names)
-    elif use_rep in adata.obsm.keys():
-        X = DataFrame(adata.obsm[use_rep], index=adata.obs_names)
-
-    if ndims_rep is not None:
-        X = X.iloc[:, :ndims_rep]
+    X = get_data(adata, use_rep, ndims_rep)
 
     if seed is not None:
         np.random.seed(seed)
@@ -750,32 +700,7 @@ def curve_epg(
     verbose: bool = True,
 ):
 
-    if use_rep not in adata.obsm.keys() and f"X_{use_rep}" in adata.obsm.keys():
-        use_rep = f"X_{use_rep}"
-
-    if (
-        (use_rep not in adata.layers.keys())
-        & (use_rep not in adata.obsm.keys())
-        & (use_rep != "X")
-    ):
-        use_rep = "X" if adata.n_vars < 50 or n_pcs == 0 else "X_pca"
-        n_pcs = None if use_rep == "X" else n_pcs
-
-    if use_rep == "X":
-        if sparse.issparse(adata.X):
-            X = DataFrame(adata.X.A, index=adata.obs_names)
-        else:
-            X = DataFrame(adata.X, index=adata.obs_names)
-    elif use_rep in adata.layers.keys():
-        if sparse.issparse(adata.layers[use_rep]):
-            X = DataFrame(adata.layers[use_rep].A, index=adata.obs_names)
-        else:
-            X = DataFrame(adata.layers[use_rep], index=adata.obs_names)
-    elif use_rep in adata.obsm.keys():
-        X = DataFrame(adata.obsm[use_rep], index=adata.obs_names)
-
-    if ndims_rep is not None:
-        X = X.iloc[:, :ndims_rep]
+    X = get_data(adata, use_rep, ndims_rep)
 
     if seed is not None:
         np.random.seed(seed)
@@ -988,7 +913,7 @@ def cleanup(
     return adata if copy else None
 
 
-def root(adata: AnnData, root: int, copy: bool = False):
+def root(adata: AnnData, root: Union[int, str], copy: bool = False):
     """\
     Define the root of the trajectory.
 
@@ -997,7 +922,7 @@ def root(adata: AnnData, root: int, copy: bool = False):
     adata
         Annotated data matrix.
     root
-        Id of the tip of the fork to be considered as a root.
+        Either an Id (int) of the tip of the fork to be considered as a root. Or a key (str) from obs (such as CytoTRACE) for automatic selection.
     copy
         Return a copy instead of writing to adata.
     Returns
@@ -1022,6 +947,18 @@ def root(adata: AnnData, root: int, copy: bool = False):
 
     graph = adata.uns["graph"]
 
+    if type(root) == str:
+        logg.info(
+            "automatic root selection using .obs['" + root + "'] values", time=False
+        )
+        avgs = list(
+            map(
+                lambda n: np.average(adata.obs[root], weights=graph["R"][:, n]),
+                range(graph["R"].shape[1]),
+            )
+        )
+        root = np.argmax(np.array(avgs))
+
     from sklearn.metrics import pairwise_distances
 
     d = 1e-6 + pairwise_distances(graph["F"].T, graph["F"].T, metric=graph["metrics"])
@@ -1081,7 +1018,11 @@ def root(adata: AnnData, root: int, copy: bool = False):
 
     adata.uns["graph"] = graph
 
-    logg.info("root selected", time=False, end=" " if settings.verbosity > 2 else "\n")
+    logg.info(
+        "node " + str(root) + " selected as a root",
+        time=False,
+        end=" " if settings.verbosity > 2 else "\n",
+    )
     logg.hint(
         "added\n"
         "    .uns['graph']['root'] selected root.\n"
@@ -1103,6 +1044,8 @@ def roots(adata: AnnData, roots, meeting, copy: bool = False):
         Annotated data matrix.
     roots
         list of tips or forks to be considered a roots.
+    meeting
+         node ID of the meeting point fo the two converging paths.
     copy
         Return a copy instead of writing to adata.
     Returns
@@ -1308,9 +1251,34 @@ def getpath(adata, root_milestone, milestones):
     return pd.concat(list(map(gatherpath, leaves)), axis=0)
 
 
-def round_base_10(x):
-    if x < 0:
-        return 0
-    elif x == 0:
-        return 10
-    return 10 ** np.ceil(np.log10(x))
+def get_data(adata, use_rep, ndims_rep):
+
+    if use_rep not in adata.obsm.keys() and f"X_{use_rep}" in adata.obsm.keys():
+        use_rep = f"X_{use_rep}"
+
+    if (
+        (use_rep not in adata.layers.keys())
+        & (use_rep not in adata.obsm.keys())
+        & (use_rep != "X")
+    ):
+        use_rep = "X" if adata.n_vars < 50 or n_pcs == 0 else "X_pca"
+        n_pcs = None if use_rep == "X" else n_pcs
+
+    if use_rep == "X":
+        ndims_rep = None
+        if sparse.issparse(adata.X):
+            X = DataFrame(adata.X.A, index=adata.obs_names)
+        else:
+            X = DataFrame(adata.X, index=adata.obs_names)
+    elif use_rep in adata.layers.keys():
+        if sparse.issparse(adata.layers[use_rep]):
+            X = DataFrame(adata.layers[use_rep].A, index=adata.obs_names)
+        else:
+            X = DataFrame(adata.layers[use_rep], index=adata.obs_names)
+    elif use_rep in adata.obsm.keys():
+        X = DataFrame(adata.obsm[use_rep], index=adata.obs_names)
+
+    if ndims_rep is not None:
+        X = X.iloc[:, :ndims_rep]
+
+    return X
