@@ -2,11 +2,13 @@ from anndata import AnnData
 import numpy as np
 import pandas as pd
 from scipy.sparse import csr_matrix
-from joblib import delayed, Parallel
+from joblib import delayed
 from tqdm import tqdm
 import sys
 import igraph
 
+from .utils import ProgressParallel
+
 from .. import logging as logg
 from .. import settings
 
@@ -58,10 +60,9 @@ def pseudotime(adata: AnnData, n_jobs: int = 1, n_map: int = 1, copy: bool = Fal
     if n_map == 1:
         df_l = [map_cells(graph, multi=False)]
     else:
-        df_l = Parallel(n_jobs=n_jobs)(
-            delayed(map_cells)(graph=graph, multi=True)
-            for m in tqdm(range(n_map), file=sys.stdout, desc="    mappings")
-        )
+        df_l = ProgressParallel(
+            n_jobs=n_jobs, total=n_map, file=sys.stdout, desc="    mappings"
+        )(delayed(map_cells)(graph=graph, multi=True) for m in range(n_map))
 
     # formatting cell projection data
 
@@ -274,11 +275,14 @@ def refine_pseudotime(
         )
         ms_data = palantir.utils.determine_multiscale_space(dm_res)
 
-    pseudotimes = Parallel(n_jobs=n_jobs)(
+    pseudotimes = ProgressParallel(
+        n_jobs=n_jobs,
+        file=sys.stdout,
+        desc="    ",
+        total=len(adata.uns["graph"]["pp_seg"].n.values),
+    )(
         delayed(palantir_on_seg)(adata, seg=s, ms_data=ms_data)
-        for s in tqdm(
-            adata.uns["graph"]["pp_seg"].n.values.astype(str), file=sys.stdout
-        )
+        for s in adata.uns["graph"]["pp_seg"].n.values.astype(str)
     )
 
     g = igraph.Graph(directed=True)
